c23456789 123456789 123456789 123456789 123456789 123456789 123456789 12
C
c     dp_pyth_wndstrp.f
c
c     This is a Python compatible Fortran routine.
c
c     This is an implementation of Algorithm adapt-DP based on DP
c     (dynamic programming) restricted to an adapting strip for
c     computing in linear time optimal diffeomorphisms for elastic
c     registration of curves. Description of Algorithm adapt-DP can
c     be found in publication "Fast Dynamic Programming for Elastic
c     Registration of Curves" that was presented at 2nd International
c     Workshop on Differential Geometry in Computer Vision and Machine
c     Learning (DIFF-CVML'16) in conjunction with Computer Vision
c     Pattern Recognition Conference (CVPR) 2016, Las Vegas, Nevada,
c     June 26-July 1, 2016.
c
c     Python compatible Fortran routine for executing Fortran
c     subroutine wind_strp_DP for computing in linear time with Python
c     approximately optimal diffeomorphism for elastic registration of
c     two curves using dynamic programming restricted to a winding
c     or adapting strip.
c     Currently wind_strp_DP is implemented so that L2 norm of
c     q1(t) - (d/dt g(t))^1/2 x q2(g(t)) is minimized by wind_strp_DP
c     with respect to g, g an orientation-preserving diffeomorphism
c     from [0,1] onto itself, where q1, q2 are q functions, i.e.,
c     square-root velocity functions from [0,1] into the real
c     line or the plane.
c
c     Compile with:
c
c     f2py -c dp_pyth_wndstrp.f -m dp_pyth_wndstrp
c
c     From Python file call as follows:
c
c     import dp_pyth_wndstrp
c
c     (g, pfly, pflp, mnrgy, ilowb, ihigh, lbnd, hbnd) =
c     dp_pyth_wndstrp.dp_pyth_wndstrp(q1, q2, t1, t2, layrs, lstrp,
c                                     iopt, lgrid, jgrid)
c     
c     Here on input q1, q2 are discretizations of q functions.
c     For the 1-d situation q1(1,t) and q2(1,t) should equal zero
c     for all t, and q1(2,t) and q2(2,t) the actual 1-d q functions.
c     For the 2-d situation (q1(1,t),q1(2,t)) is q function for first
c     curve, and (q2(1,t),q2(2,t)) is q function for second curve.
c
c     On input t1, t2 are discrete partitions of [0,1] not
c     necessarily uniform or equal to each other. t1 is the domain
c     of q1 and t2 is the domain of q2.
c
c     On input layrs is the number of layers of grid points in local
c     square-shaped neighborhood associated with current grid point
c     during dynamic programming process. Grid points in local
c     neighborhood are the only grid points below current grid point
c     that are considered as possible neighbors of current grid point
c     in the graph of any piece-wise linear homeomorphism from [0,1]
c     onto itself that approximates diffeomorphism from [0,1] onto
c     itself during the dynamic programming process. 
c
c     On input lstrp is a number that determines width of winding
c     strip. Number of grid points across winding strip is
c     approximately lstrp times the square root of 2.
c
c     On input iopt is a number that equals 1 if 1-d situation is being
c     considered, and equals 2 otherwise (2-d situation).
c
c     On input lgrid, jgrid are the lengths of q1, t1, and q2, t2,
c     respectively, i.e., lgrid is the number of points on first curve,
c     jgrid the number on second curve.
c
c     On output g is the optimal diffeomorphism approximation, i.e.,
c     grid points that define optimal piece-wise linear homeomorphism
c     from planar point (0,0) to planar point (1,1).
c
c     On ouput pfly is percentage of vertices (grid points) of
c     approximately optimal homeomorphism that are separated from
c     immediate vertex below by a number of layers of grid points
c     equal or larger than layrs.
c
c     On ouput pflp is percentage of vertices (grid points) of
c     approximately optimal homeomorphism that are on boundary
c     of final strip.
c
c     On output mnrgy is approximately optimal energy computed by
c     dynamic programming procedure, i.e., approximation of optimal
c     energy that would be the minimum of L2 norm of
c     q1(t) - (d/dt g(t))^1/2 x q2(g(t)).
c
c     On output ilowb is the number of grid points on the lower
c     boundary of last strip.
c
c     On output ihigh is the number of grid points on the upper
c     boundary of last strip.
c
c     On output lbnd is set of grid points in their planar
c     coordinates on the lower boundary of last strip.
c
c     On output hbnd is set of grid points in their planar
c     coordinates on the upper boundary of last strip.
c
c*******************************************************************
c
c
c     Compile with:
c
c     f2py -c dp_pyth_wndstrp.f -m dp_pyth_wndstrp
c
c     From Python file call as follows:
c
c     import dp_pyth_wndstrp
c
c     (g, pfly, pflp, mnrgy, ilowb, ihigh, lbnd, hbnd) =
c     dp_pyth_wndstrp.dp_pyth_wndstrp(q1, q2, t1, t2, layrs, lstrp,
c                                     iopt, lgrid, jgrid)
c     
      subroutine dp_pyth_wndstrp(q1, q2, t1, t2, layrs, lstrp, iopt,
     *                           lgrid, jgrid, g, pfly, pflp, mnrgy,
     *                           ilowb, ihigh, lbnd, hbnd)
c
c     input variables:
c
c     q1 - two dimensional double precision array
c     q2 - two dimensional double precision array
c     t1 - one dimensional double precision array
c     t2 - one dimensional double precision array
c     layrs - integer scalar
c     lstrp - integer scalar
c     iopt - integer scalar
c     lgrid - integer scalar
c     jgrid - integer scalar
c
c     output variables:
c
c     g - one dimensional double precision array
c     pfly  - double precision scalar
c     pflp  - double precision scalar
c     mnrgy - double precision scalar
c     ilowb - integer scalar
c     ihigh - integer scalar
c     lbnd - two dimensional double precision array
c     hbnd - two dimensional double precision array
c
C-------------------------------------------------------------------
C
      integer cmax, nmax, lmax, gmax, wmax, hmax, vmax
      parameter (cmax = 5, nmax = 50001, lmax = cmax*nmax,
     *         gmax = 2*nmax, wmax = 1000, hmax = 10001, vmax = 10000)
C
      integer j, lgrid, jgrid, mgrid, layrs, lstrp, iopt
      double precision q1(2,lgrid), q2(2,jgrid), g(lgrid)
      double precision t1(lgrid), t2(jgrid)
      double precision h2, pfly, pflp, mnrgy
c
      double precision qm1(lmax), qm2(lmax), s2(lmax)
      double precision lbnd(2,gmax), hbnd(2,gmax), wk(2,nmax)
      double precision s(nmax), r(nmax)
      double precision t1m(nmax), t2m(nmax)
      double precision bndst(wmax,nmax), bltst(hmax,vmax), bntmp(wmax)
      integer instr(wmax,nmax), ilstp(hmax,vmax), iwstr(nmax)
      integer iwini(nmax), iwlst(nmax), iwall(nmax), intmp(wmax)
      integer igig(2,gmax), idig(2,gmax), iretrn, ilowb, ihigh
      integer iar(nmax), ibr(nmax), ian(nmax), ibn(nmax)
      integer iat(nmax), ibt(nmax)
c
c     intents (necessary for Python to understand what is input and
c     what is output):
c
cf2py intent(in) :: q1
cf2py intent(in) :: q2
cf2py intent(in) :: t1
cf2py intent(in) :: t2
cf2py intent(in) :: layrs
cf2py intent(in) :: lstrp
cf2py intent(in) :: iopt
cf2py intent(in) :: lgrid
cf2py intent(in) :: jgrid
cf2py intent(out) :: g
cf2py intent(out) :: pfly
cf2py intent(out) :: pflp
cf2py intent(out) :: mnrgy
cf2py intent(out) :: ilowb
cf2py intent(out) :: ihigh
cf2py intent(out) :: lbnd
cf2py intent(out) :: hbnd
c
C-------------------------------------------------------------------
c     write(*,*)'lgrid=',lgrid,' jgrid=',jgrid,' layrs=',layrs,
c    *          ' lstrp=',lstrp,' iopt=',iopt
c     write(*,*)(q1(1,j),q1(2,j),j=1,10)
c     write(*,*)(q2(1,j),q2(2,j),j=1,10)
c
      mgrid = cmax*(jgrid-1) + 1
c
      if(lgrid.gt.nmax .or. jgrid.gt.nmax) then
         write(*,*)'Parameter nmax exceeded by lgrid, jgrid.'
         stop
      endif
      if(mgrid.gt.lmax) then
         write(*,*)'Parameter lmax exceeded.'
         stop
      endif
c
c-------------------------------------------------------------
c
c     initialize arrays s2
c
      h2 = 1.0d0/dble(mgrid-1)
      s2(1) = 0.0d0
      do 650 j = 2, mgrid-1
         s2(j) = dble(j-1)*h2
  650 continue
      s2(mgrid) = 1.0d0
c
      if(iopt.ne.1 .and. iopt.ne.2) then
         write(*,*)'Wrong option.'
         stop
      endif
c
      iretrn = 0
c
C     Call computational subroutine
c
      call wind_strp_DP(bndst, bltst, bntmp, instr, ilstp, iwstr,
     *            iwini, iwlst, iwall, intmp, q1, q2, qm1, qm2, lbnd,
     *            hbnd, wk, t1, t2, t1m, t2m, s2, s, r, g, igig, idig,
     *            iar, ibr, ian, ibn, iat, ibt, iretrn,
     *            ilowb, ihigh, lgrid, jgrid, mgrid,
     *            cmax, nmax, lmax, gmax, wmax, hmax, vmax,
     *            layrs, lstrp, iopt, mnrgy, pfly, pflp)
c
      if(iretrn.eq.1) then
         write(*,*)'Termination: hmax <= wmax - should be hmax > wmax'
         stop
      elseif(iretrn.eq.2) then
         write(*,*)'Termination: vmax exceeded ',
     *             '- still keep hmax*vmax <= O(N)'
         stop
      elseif(iretrn.eq.3) then
         write(*,*)'Termination: hmax exceeded ',
     *             '- still keep hmax*vmax <= O(N)'
         stop
      elseif(iretrn.eq.4) then
         write(*,*)'Termination: wmax exceeded ',
     *             '- still keep wmax*nmax <= O(N)'
         stop
      endif
c
      if(ilowb.gt.gmax .or. ihigh.gt.gmax) then
         write(*,*)'On output ilowb > gmax or ihigh > gmax.'
         stop
      endif
c
      return
      end
c
*WIND_STRP_DP
c
c     subroutine wind_strp_DP
c
c     for finding, using dynamic programming, approximation of
c     orientation-preserving diffeomorphism g from the interval
c     [0,1] onto itself that minimizes the L2 norm of q1(t) -
c     (d/dt g(t))^1/2 x q2(g(t)), where q1 and q2 are q functions,
c     i.e., square-root velocity functions, in L2 space, of norm 1,
c     from [0,1] into the real line or the plane.
c
c*******************************************************************
c
c     Options:
c
c     iopt = 1 
c     
c     Used for warping functions in the plane. Accordingly q1 and q2
c     although defined as 2-d arrays are actually understood to be
c     1-d arrays for the purpose of representing q functions in the
c     xy plane whose domain is the [0,1] interval in the x line, and
c     whose range is the y line. Accordingly q1(1,t) and q2(1,t) equal
c     zero for all t, and q1(2,t) and q2(2,t) represent q functions.
c
c     iopt = 2
c
c     Used for computing geodesic distances between shapes in the
c     plane. Accordingly (q1(1,t),q1(2,t)) is q function for first
c     curve, and (q2(1,t),q2(2,t)) is q function for second curve.
c
c*******************************************************************
c
c    Author: Javier Bernal
c            National Institute of Standards and Technology (NIST)
c
c     Disclaimer:
c
c     This software was developed at the National Institute of Standards
c     and Technology by employees of the Federal Government in the
c     course of their official duties. Pursuant to title 17 Section 105
c     of the United States Code this software is not subject to
c     copyright protection and is in the public domain. This software is
c     experimental. NIST assumes no responsibility whatsoever for its
c     use by other parties, and makes no guarantees, expressed or
c     implied, about its quality, reliability, or any other
c     characteristic. We would appreciate acknowledgement if the
c     software is used.
c
      subroutine wind_strp_DP(bndst, bltst, bntmp, instr, ilstp, iwstr,
     *                  iwini, iwlst, iwall, intmp, q1, q2, qm1, qm2,
     *                  qc, del, wk, t1, t2, t1m, t2m, s2, s, r, g,
     *                  igig, idig, iar, ibr, ian, ibn, iat, ibt,
     *                  iretrn, ilowb, ihigh, lgrid, jgrid, mgrid,
     *                  cmax, nmax, lmax, gmax, wmax, hmax, vmax,
     *                  layrs, lstrp, iopt, mnrgy, pfly, pflp)
c
      integer cmax, nmax, lmax, gmax, wmax, hmax, vmax
c
      double precision qm1(*), qm2(*), s2(*)
      double precision q1(2,*), q2(2,*), qc(2,*), del(2,*)
      double precision s(*), r(*), g(*)
      double precision t1(*), t2(*), t1m(*), t2m(*)
      double precision bndst(wmax,*), bltst(hmax,*)
      double precision bntmp(*), bntnp, bltmp
      integer instr(wmax,*), ilstp(hmax,*), iwstr(*)
      integer iwini(*), iwlst(*), iwall(*), intmp(*), intnp
      integer igig(2,*), idig(2,*)
      integer iar(*), ibr(*), ian(*), ibn(*)
      integer iat(*), ibt(*)
c
      double precision mnrgy, pfly, pflp
      double precision bdust, area, ginc, vx, vy, t10, t20, dot
      integer ifly, iflp, iflw, iopt, iretrn, ilowb, ihigh
      integer lgrid, jgrid, mgrid, ldiag, lgiag, ispc, istg
      integer lstrp, layrs, lstri, lstrj, layri, layrj
      integer ifrt, jfrt, i, j, k, l, ii, jj
      integer ix, jx, iw, jw, iv, jv, jg
      integer ifull, jfull, ifill, jfill
c     INTEGER IPASS
c
      integer ic(2), incfd, nwk, ierr
      double precision vc(2), wk(2,*)
      logical skip
      parameter (incfd = 2)
c
c     lgrid = i dimension of grid or length of q1
c     jgrid = j dimension of grid or length of q2
c
c     lstrp = maximum number of acceptable grid points from each
c             grid point on diagonal to the left and below
c
c     layrs = number of layers around each grid point to test
c
      ldiag = 0
      jv = 0
c
c     check dimensions, etc.
c
      if(iopt.ne.1 .and. iopt.ne.2) stop 80
      if(hmax.le.wmax) then
         iretrn = 1
         return
      endif
      if(lgrid.lt.4 .or. jgrid.lt.4) stop 90
      if(lgrid.gt.nmax .or. jgrid.gt.nmax) stop 100
      if(mgrid.gt.lmax) stop 110
      if(mgrid-1.lt.jgrid-1) stop 120
c
c     lstrp = pripr*lgrid
c     lstrp = 16
c     WRITE(*,*)'LSTRP=',LSTRP
      if(lstrp.lt.0) stop 130
c
c     WRITE(*,*)'LAYRS=',LAYRS
      if(layrs.lt.1) stop 140
c
c     initialize for spline
c
      nwk = 2*jgrid
      ic(1) = 0
      ic(2) = 0
      vc(1) = 0.0d0
      vc(2) = 0.0d0
      skip = .false.
c
      if(cmax.gt.1) then
         if(iopt.eq.1) go to 10
         call pchsp(ic, vc, jgrid, t2, q2, del, incfd, wk, nwk, ierr)
         call pchfe(jgrid, t2, q2, del, incfd, skip, mgrid, s2, qm1,
     *   ierr)
c
   10    continue
         call pchsp(ic, vc, jgrid, t2, q2(2,1), del(2,1), incfd, wk,
     *           nwk, ierr)
         call pchfe(jgrid, t2, q2(2,1), del(2,1), incfd, skip,
     *           mgrid, s2, qm2, ierr)
      else
         if(jgrid.ne.mgrid) stop 150
         do 15 i = 1, mgrid
            qm1(i) = q2(1,i)
            qm2(i) = q2(2,i)
   15    continue
      endif
c
c     initialize for connecting arrays
c
      iat(1) = 1
      do 50 i = 2, lgrid
         iat(i) = 0
         iar(i) = 0
         ian(i) = 0
         if(t1(i).le.t1(i-1)) stop 152
   50 continue
c
      ibt(1) = 1
      iwall(1) = 1
      do 55 j = 2, jgrid
         ibt(j) = 0
         ibr(j) = 0
         ibn(j) = 0
         if(t2(j).le.t2(j-1)) stop 154
         iwall(j) = 0
   55 continue
c
c     set initial supports
c
      ii = idnint(dble(lgrid+1)/3.0d0)
      iar(ii) = 1
      ian(1) = ii
      i = idnint(dble(2*(lgrid+1))/3.0d0)
      iar(i) = ii
      ian(ii) = i
      iar(lgrid) = i
      ian(i) = lgrid
      iar(1) = iar(lgrid)
      ian(lgrid) = lgrid+1
c
      jj = idnint(dble(jgrid+1)/3.0d0)
      ibr(jj) = 1
      ibn(1) = jj
      j = idnint(dble(2*(jgrid+1))/3.0d0)
      ibr(j) = jj
      ibn(jj) = j
      ibr(jgrid) = j
      ibn(j) = jgrid
      ibr(1) = ibr(jgrid)
      ibn(jgrid) = jgrid+1
c
c     set initial diffeomorphism
c
      lgiag = 2
      if(lgiag.gt.gmax) stop 156
      igig(1,1) = 1
      igig(2,1) = 1
      igig(1,2) = lgrid
      igig(2,2) = jgrid
c
c     WRITE(*,*)'IBR='
c     WRITE(*,85)(IBR(J),J=1,JGRID)
c  85 FORMAT(10i5,1x)
c
c     compute midpoints between consecutive grid points
c
c     IPASS = 0
      ifull = 0
      jfull = 0
      IFILL = 0
      JFILL = 0
  100 continue
c     IPASS = IPASS+1
c     WRITE(*,*)'IPASS=',IPASS
      i = ian(1)
      if(i.eq.0) stop 160
  120 continue
      ii = ian(i)
      if(ii.eq.lgrid) go to 125
      if(ii.eq.0) stop 170
      t1m(i) = (t1(ii)+t1(i))/2.0d0
      i = ii
      go to 120
  125 continue
      t1m(i) = 1.0d0
c
      j = ibn(1)
      if(j.eq.0) stop 180
      k = 2
      iwall(j) = 2
  130 continue
      jj = ibn(j)
      if(jj.eq.jgrid) go to 135
      if(jj.eq.0) stop 190
      t2m(j) = (t2(jj)+t2(j))/2.0d0
      j = jj
      k = k+1
      iwall(j) = k
      go to 130
  135 continue
      t2m(j) = 1.0d0
      k = k+1
      iwall(jgrid) = k
c
c     identify grid points close to current diffeomorphism
c
      ii = iar(lgrid)
      jj = ibr(jgrid)
c     IF(IPASS.EQ.8)WRITE(*,*)'Termination grid point II JJ=', II, JJ
      if(ii.eq.0 .or. jj.eq.0) stop 200
      do 170 l = 2, lgiag
c        IF(IPASS.EQ.8)WRITE(*,*)'Current piece of diffeomorphism =',l
         ix = igig(1,l)
         jx = igig(2,l)
c        IF(IPASS.EQ.8)WRITE(*,*)'IX JX = ',IX,JX
         if(l.eq.2) then
            i = ian(1)
            j = ibn(1)
            if(i.eq.0 .or. j.eq.0) stop 210
            ldiag = 1
            idig(1,1) = i
            idig(2,1) = j
            if(i.eq.ix .and. j.eq.jx) go to 170
            vx = t2(1)-t2(jx)
            vy = t1(ix)-t1(1)
            t10 = t1(1)
            t20 = t2(1)
         else
            i = igig(1,l-1)
            j = igig(2,l-1)
            vx = t2(j)-t2(jx)
            vy = t1(ix)-t1(i)
            t10 = t1(i)
            t20 = t2(j)
         endif
c        IF(IPASS.EQ.8)WRITE(*,*)'Initial grid point I J =',I,J
  140    continue
         dot = vx*(t1m(i)-t10) + vy*(t2m(j)-t20)
         if(dot.lt.0.0d0) then
            j = ibn(j)
         elseif(dot.gt.0.0d0) then
            i = ian(i)
         else
            i = ian(i)
            j = ibn(j)
         endif
         if(i.eq.0 .or. j.eq.0) stop 220
         ldiag = ldiag+1
         if(ldiag.gt.gmax) stop 230
c        IF(IPASS.EQ.8)WRITE(*,*)'Current I J =',I,J
         idig(1,ldiag) = i
         idig(2,ldiag) = j
         if(i.eq.ii .and. j.eq.jj) go to 175
         if(i.gt.ix .or. j.gt.jx) stop 240
         if(i.lt.ix .or. j.lt.jx) go to 140
  170 continue
  175 continue
      ldiag = ldiag+1
      if(ldiag.gt.gmax) stop 250
      idig(1,ldiag) = lgrid
      idig(2,ldiag) = jgrid
c     IF(IPASS.EQ.8)WRITE(*,*)'Final grid point I J =',LGRID,JGRID
c
c     visit every acceptable grid point
c
      ifrt = idig(1,1)
      jfrt = idig(2,1)
      iwini(1) = 1
      iwlst(1) = 1
c
      iv = 0
      do 400 l = 1, ldiag
         i = idig(1,l)
         j = idig(2,l)
         if(iar(i).eq.0 .or. ibr(j).eq.0) stop 260
         iat(i) = j
         ibt(j) = i
         jw = iwall(j)
         if(jw.eq.0) stop 272
         ispc = 1
         if(l.eq.1) go to 200
         if(i.ne.idig(1,l-1)) go to 200
         if(jw.ne.iwlst(i)+1) stop 274
         iwlst(i) = jw
         iw = jw-iwini(i)+1
         ispc = iwstr(i)
         if(ispc.lt.0) go to 183
         if(ispc.eq.0) stop 276
         if(iw.le.wmax) go to 210
         iv = iv+1
         if(iv.gt.vmax) then
            iretrn = 2
            return
         endif
         do 182 ix = 1, iw-1
            bltst(ix,iv) = bndst(ix,i)
            ilstp(ix,iv) = instr(ix,i)
  182    continue
         ispc = -iv
         iwstr(i) = ispc
  183    continue
         if(iw.gt.hmax) then
            iretrn = 3
            return
         endif
         jv = -ispc
         go to 210
  200    continue
         iwstr(i) = 1
         iwini(i) = jw
         iwlst(i) = jw
         iw = 1
  210    continue
         bntmp(1) = 0.0d0
         intmp(1) = -1
         if(i.eq.ifrt .or. j.eq.jfrt) then
            ii = 1
            jj = 1
            call intopt(q1, q2, qm1, qm2, qc, t1, t2, s, r, iar,
     *                  area, mgrid, ii, jj, i, j, iopt)
            bntmp(1) = area
            intmp(1) = 1
            go to 240
         endif
         jj = j
         do 230 layrj = 1, layrs
            jj = ibr(jj)
            if(jj.eq.1) go to 235
            if(ibr(jj).eq.0) stop 280
            jx = iwall(jj)
            if(jx.eq.0) stop 282
            ii = i
            do 220 layri = 1, layrs
               ii = iar(ii)
               if(ii.eq.1) go to 230
               if(iar(ii).eq.0) stop 290
               if(jx.lt.iwini(ii) .or. jx.gt.iwlst(ii)) go to 220
               ix = jx-iwini(ii)+1
               istg = iwstr(ii)
               if(istg.gt.0) then
                  bltmp = bndst(ix,ii)
               elseif(istg.lt.0) then
                  jg = -istg
                  bltmp = bltst(ix,jg)
               else
                  stop 292
               endif
               if(intmp(1).ne.-1 .and. bntmp(1).le.bltmp)
     *            go to 220
               call intopt(q1, q2, qm1, qm2, qc, t1, t2, s, r, iar,
     *                     area, mgrid, ii, jj, i, j, iopt)
               bdust = area + bltmp
               if(intmp(1).eq.-1 .or.
     *            (intmp(1).ne.-1.and.bdust.lt.bntmp(1))) then
                  bntmp(1) = bdust
                  intmp(1) = (jj-1)*lgrid + ii
               endif
  220       continue
  230    continue
  235    continue
         if(intmp(1).ne.-1) go to 238
         if(l.eq.1) stop 294
         if(j.eq.idig(2,l-1)) then
            jj = ibr(j)
            if(ibr(jj).eq.0) stop 300
            ii = ibt(jj)
            if(ii.eq.0) stop 310
            jx = iwall(jj)
            if(jx.eq.0) stop 312
            if(jx.lt.iwini(ii) .or. jx.gt.iwlst(ii)) stop 320
            ix = jx-iwini(ii)+1
            istg = iwstr(ii)
            if(istg.gt.0) then
               bltmp = bndst(ix,ii)
            elseif(istg.lt.0) then
               jg = -istg
               bltmp = bltst(ix,jg)
            else
               stop 325
            endif
            call intopt(q1, q2, qm1, qm2, qc, t1, t2, s, r, iar,
     *                  area, mgrid, ii, jj, i, j, iopt)
            bntmp(1) = area + bltmp
            intmp(1) = (jj-1)*lgrid + ii
c
         else
            ii = iar(i)
            if(iar(ii).eq.0) stop 330
            jj = iat(ii)
            if(jj.eq.0) stop 340
            jx = iwall(jj)
            if(jx.eq.0) stop 342
            if(jx.lt.iwini(ii) .or. jx.gt.iwlst(ii)) stop 344
            ix = jx-iwini(ii)+1
            istg = iwstr(ii)
            if(istg.gt.0) then
               bltmp = bndst(ix,ii)
            elseif(istg.lt.0) then
               jg = -istg
               bltmp = bltst(ix,jg)
            else
               stop 350
            endif
            call intopt(q1, q2, qm1, qm2, qc, t1, t2, s, r, iar,
     *                  area, mgrid, ii, jj, i, j, iopt)
            bntmp(1) = area + bltmp
            intmp(1) = (jj-1)*lgrid + ii
         endif
  238    continue
c
  240    continue
         if(ispc.gt.0) then
            bndst(iw,i) = bntmp(1)
            instr(iw,i) = intmp(1)
         else
            bltst(iw,jv) = bntmp(1)
            ilstp(iw,jv) = intmp(1)
         endif
c        WRITE(*,*)'IW I INSTR(IW,I) =',IW,I,INSTR(IW,I)
         if((i.eq.ifrt.and.j.eq.jfrt) .or. l.eq.ldiag) go to 400
         if(lstrp.eq.0) go to 400
         if(l.eq.1) stop 352
         if(j.eq.idig(2,l-1)) go to 280
         if(ibr(j).eq.1) stop 354
c
         jw = iwall(j)
         if(jw.eq.0) stop 356
         do 260 lstri = 1, lstrp
            i = iar(i)
            if(i.eq.1) go to 280
            if(iar(i).eq.0) stop 360
            iat(i) = j
            if(jw.ne.iwlst(i)+1) stop 370
            iwlst(i) = jw
            iw = jw-iwini(i)+1
            ispc = iwstr(i)
            if(ispc.lt.0) go to 243
            if(ispc.eq.0) stop 372
            if(iw.le.wmax) go to 244
            iv = iv+1
            if(iv.gt.vmax) then
               iretrn = 2
               return
            endif
            do 242 ix = 1, iw-1
               bltst(ix,iv) = bndst(ix,i)
               ilstp(ix,iv) = instr(ix,i)
  242       continue
            ispc = -iv
            iwstr(i) = ispc
  243       continue
            if(iw.gt.hmax) then
               iretrn = 3
               return
            endif
            jv = -ispc
  244       continue
            bntnp = 0.0d0
            intnp = -1
            if(i.eq.ifrt .or. j.eq.jfrt) then
               ii = 1
               jj = 1
               call intopt(q1, q2, qm1, qm2, qc, t1, t2, s, r, iar,
     *                     area, mgrid, ii, jj, i, j, iopt)
               if(ispc.gt.0) then
                  bndst(iw,i) = area
                  instr(iw,i) = 1
               else
                  bltst(iw,jv) = area
                  ilstp(iw,jv) = 1
               endif
               if(i.eq.ifrt) go to 280
               go to 260
            endif
            jj = j
            do 250 layrj = 1, layrs
               jj = ibr(jj)
               if(jj.eq.1) go to 255
               if(ibr(jj).eq.0) stop 380
               jx = iwall(jj)
               if(jx.eq.0) stop 382
               ii = i
               do 245 layri = 1, layrs
                  ii = iar(ii)
                  if(ii.eq.1) go to 250
                  if(iar(ii).eq.0) stop 390
                  if(jx.lt.iwini(ii) .or. jx.gt.iwlst(ii)) go to 245
                  ix = jx-iwini(ii)+1
                  istg = iwstr(ii)
                  if(istg.gt.0) then
                     bltmp = bndst(ix,ii)
                  elseif(istg.lt.0) then
                     jg = -istg
                     bltmp = bltst(ix,jg)
                  else
                     stop 392
                  endif
                  if(intnp.ne.-1 .and. bntnp.le.bltmp)
     *               go to 245
                  call intopt(q1, q2, qm1, qm2, qc, t1, t2, s, r, iar,
     *                        area, mgrid, ii, jj, i, j, iopt)
                  bdust = area + bltmp
                  if(intnp.eq.-1 .or.
     *               (intnp.ne.-1.and.bdust.lt.bntnp)) then
                     bntnp = bdust
                     intnp = (jj-1)*lgrid + ii
                  endif
  245          continue
  250       continue
  255       continue
            if(intnp.eq.-1) then
               ii = iar(i)
               if(iar(ii).eq.0) stop 400
               jj = iat(ii)
               if(jj.eq.0) stop 410
               jx = iwall(jj)
               if(jx.eq.0) stop 412
               if(jx.lt.iwini(ii) .or. jx.gt.iwlst(ii)) stop 420
               ix = jx-iwini(ii)+1
               istg = iwstr(ii)
               if(istg.gt.0) then
                  bltmp = bndst(ix,ii)
               elseif(istg.lt.0) then
                  jg = -istg
                  bltmp = bltst(ix,jg)
               else
                  stop 422
               endif
               call intopt(q1, q2, qm1, qm2, qc, t1, t2, s, r, iar,
     *                     area, mgrid, ii, jj, i, j, iopt)
               bntnp = area + bltmp
               intnp = (jj-1)*lgrid + ii
            endif
            if(ispc.gt.0) then
               bndst(iw,i) = bntnp
               instr(iw,i) = intnp
            else
               bltst(iw,jv) = bntnp
               ilstp(iw,jv) = intnp
            endif
  260    continue
c
  280    continue
         i = idig(1,l)
         if(l.eq.1) stop 424
         if(i.eq.idig(1,l-1)) go to 400
         if(iar(i).eq.1) stop 426
c
         iw = 1
         do 360 lstrj = 1, lstrp
            j = ibr(j)
            if(j.eq.1) go to 380
            if(ibr(j).eq.0) stop 430
            ibt(j) = i
            jw = iwall(j)
            if(jw.eq.0) stop 442
            iwini(i) = jw
            iw = iw+1
            if(iw.gt.wmax) then
               iretrn = 4
               return
            endif
            bntmp(iw) = 0.0d0
            intmp(iw) = -1
            if(j.eq.jfrt .or. i.eq.ifrt) then
               ii = 1
               jj = 1
               call intopt(q1, q2, qm1, qm2, qc, t1, t2, s, r, iar,
     *                     area, mgrid, ii, jj, i, j, iopt)
               bntmp(iw) = area
               intmp(iw) = 1
               if(j.eq.jfrt) go to 380
               go to 360
            endif
            jj = j
            do 350 layrj = 1, layrs
               jj = ibr(jj)
               if(jj.eq.1) go to 355
               if(ibr(jj).eq.0) stop 450
               jx = iwall(jj)
               if(jx.eq.0) stop 452
               ii = i
               do 345 layri = 1, layrs
                  ii = iar(ii)
                  if(ii.eq.1) go to 350
                  if(iar(ii).eq.0) stop 460
                  if(jx.lt.iwini(ii) .or. jx.gt.iwlst(ii)) go to 345
                  ix = jx-iwini(ii)+1
                  istg = iwstr(ii)
                  if(istg.gt.0) then
                     bltmp = bndst(ix,ii)
                  elseif(istg.lt.0) then
                     jg = -istg
                     bltmp = bltst(ix,jg)
                  else
                     stop 462
                  endif
                  if(intmp(iw).ne.-1 .and. bntmp(iw).le.bltmp)
     *               go to 345
                  call intopt(q1, q2, qm1, qm2, qc, t1, t2, s, r, iar,
     *                        area, mgrid, ii, jj, i, j, iopt)
                  bdust = area + bltmp
                  if(intmp(iw).eq.-1 .or.
     *               (intmp(iw).ne.-1.and.bdust.lt.bntmp(iw))) then
                     bntmp(iw) = bdust
                     intmp(iw) = (jj-1)*lgrid + ii
                  endif
  345          continue
  350       continue
  355       continue
            if(intmp(iw).eq.-1) then
               jj = ibr(j)
               if(ibr(jj).eq.0) stop 470
               ii = ibt(jj)
               if(ii.eq.0) stop 480
               jx = iwall(jj)
               if(jx.eq.0) stop 482
               if(jx.lt.iwini(ii) .or. jx.gt.iwlst(ii)) stop 490
               ix = jx-iwini(ii)+1
               istg = iwstr(ii)
               if(istg.gt.0) then
                  bltmp = bndst(ix,ii)
               elseif(istg.lt.0) then
                  jg = -istg
                  bltmp = bltst(ix,jg)
               else
                  stop 492
               endif
               call intopt(q1, q2, qm1, qm2, qc, t1, t2, s, r, iar,
     *                     area, mgrid, ii, jj, i, j, iopt)
               bntmp(iw) = area + bltmp
               intmp(iw) = (jj-1)*lgrid + ii
            endif
  360    continue
  380    continue
         if(iw.eq.1) go to 400
         do 390 ix = 1, iw
            jx = iw-ix+1
            bndst(ix,i) = bntmp(jx)
            instr(ix,i) = intmp(jx)
  390    continue
c
  400 continue
c     WRITE(*,*)'IV=',IV
c     MNRGY = BNDST(1,LGRID)
c     WRITE(*,*)'MNRGY=',MNRGY,' IFILL=',IFILL,' JFILL=',JFILL
c
c     starting at grid point (lgrid,jgrid) find graph of optimal
c     diffeormorphism by moving all the way to grid point (1,1)
c
      i = lgrid
      j = jgrid
      ldiag = 1
      idig(1,1) = i
      idig(2,1) = j
  700 continue
      jw = iwall(j)
      if(jw.eq.0) stop 502
      if(jw.lt.iwini(i) .or. jw.gt.iwlst(i)) stop 504
      iw = jw-iwini(i)+1
      ispc = iwstr(i)
      if(ispc.gt.0) then
         k = instr(iw,i)
      elseif(ispc.lt.0) then
         jv = -ispc
         k = ilstp(iw,jv)
      else
         stop 506
      endif
      jj = k/lgrid
      ii = k-(jj*lgrid)
      if(ii.le.0) stop 510
      jj = jj+1
      if(jj.ge.jgrid) stop 520
      if(ii.ge.i .or. jj.ge.j) stop 530
      if(iar(ii).eq.0 .or. ibr(jj).eq.0) stop 540
      ldiag = ldiag+1
      if(ldiag.gt.gmax) stop 550
      idig(1,ldiag) = ii
      idig(2,ldiag) = jj
c
      if(ii.eq.1.or.jj.eq.1) go to 800
      i = ii
      j = jj
      go to 700
  800 continue
      if(ii.ne.1.or.jj.ne.1) stop 560
c
c     set current diffeomorphism
c
      lgiag = ldiag
      do 950 l = 1, lgiag
         igig(1,l) = idig(1,ldiag-l+1)
         igig(2,l) = idig(2,ldiag-l+1)
  950 continue
c
c     set next supports
c
      if(ifull.eq.1 .and. jfull.eq.1) go to 1100
      if(ifull.eq.1) go to 1010
      ifill = 0
      i = 1
  960 continue
      ii = ian(i)
      if(ii.eq.0) stop 580
      if(i.eq.ii-1) go to 980
      ifill = ifill+1
      ix = (ii+i)/2
      if(iar(ix).ne.0 .or. ian(ix).ne.0) stop 590
      iar(ix) = i
      ian(ix) = ii
      iar(ii) = ix
      ian(i) = ix
  980 continue
      if(ii.eq.lgrid) go to 1000
      i = ii
      go to 960
 1000 continue
      iar(1) = iar(lgrid)
      if(ifill.eq.0) ifull = 1
c     WRITE(*,*)'IFILL=',IFILL
c
 1010 continue
      if(jfull.eq.1) go to 100
      jfill = 0
      j = 1
 1020 continue
      jj = ibn(j)
      if(jj.eq.0) stop 600
      if(j.eq.jj-1) go to 1030
      jfill = jfill+1
      jx = (jj+j)/2
      if(ibr(jx).ne.0 .or. ibn(jx).ne.0) stop 604
      ibr(jx) = j
      ibn(jx) = jj
      ibr(jj) = jx
      ibn(j) = jx
 1030 continue
      if(jj.eq.jgrid) go to 1050
      j = jj
      go to 1020
 1050 continue
      ibr(1) = ibr(jgrid)
      if(jfill.eq.0) jfull = 1
c     WRITE(*,*)'JFILL=',JFILL
      go to 100
 1100 continue
c
c     identify border of last strip
c
      ilowb = 0
      do 1140 i = 2, lgrid-1
         jx = iwini(i)
         ispc = iwstr(i)
         if(ispc.lt.0) jv = -ispc
         ii = i+1
         jw = iwini(ii)
         iw = jw - jx + 1
         if(iw.le.0) stop 610
         if(i.eq.lgrid-1) iw = 1
         if(jx.eq.2 .and. iw.eq.1) go to 1140
         do 1120 ix = 1, iw
            ilowb = ilowb+1
            if(ilowb.gt.gmax) stop 612
            j = ix+jx-1
            qc(1,ilowb) = t1(i)
            qc(2,ilowb) = t2(j)
            if(ispc.gt.0) then
               if(instr(ix,i).lt.0) stop 614
               instr(ix,i) = -instr(ix,i)
            else
               if(ilstp(ix,jv).lt.0) stop 616
               ilstp(ix,jv) = -ilstp(ix,jv)
            endif
 1120    continue
 1140 continue
c
      ihigh = 0
      do 1180 i = 2, lgrid-1
         jx = iwlst(i)
         if(i.eq.2 .and. jx.eq.jgrid-1) go to 1190
         ispc = iwstr(i)
         if(ispc.lt.0) jv = -ispc
         ii = i-1
         jw = iwlst(ii)
         iw = jx-jw+1
         if(iw.le.0) stop 618
         if(i.eq.2) iw = 1
         k = iwini(i)
         jg = jx-k+1
         jj = jg+1-iw
         do 1160 ix = jj, jg
            ihigh = ihigh+1
            if(ihigh.gt.gmax) stop 620
            j = ix+k-1
            del(1,ihigh) = t1(i)
            del(2,ihigh) = t2(j)
            if(ispc.gt.0) then
               if(instr(ix,i).lt.0) stop 622
               instr(ix,i) = -instr(ix,i)
            else
               if(ilstp(ix,jv).lt.0) stop 624
               ilstp(ix,jv) = -ilstp(ix,jv)
            endif
 1160    continue
         if(jx.eq.jgrid-1) go to 1190
 1180 continue
 1190 continue
c
c     starting at grid point (lgrid,jgrid) identify optimal
c     diffeormorphism by moving all the way to grid point (1,1)
c
      mnrgy = bndst(1,lgrid)
      ifly = 0
      iflp = 0
      iflw = 0
      ldiag = 1
      i = lgrid
      j = jgrid
      g(i) = 1.0d0
 1200 continue
      ldiag = ldiag+1
      ii = idig(1,ldiag)
      jj = idig(2,ldiag)
      ginc = (t2(j)-t2(jj))/(t1(i)-t1(ii))
      g(ii) = t2(jj)
c
      ix = i
 1240 continue
      ix = ix-1
      if(ix.eq.ii) go to 1260
      g(ix) = g(ii) + ginc*(t1(ix)-t1(ii))
      go to 1240
 1260 continue
c
      iflw = iflw+1
      if(i-ii.ge.layrs .or. j-jj.ge.layrs) ifly = ifly+1
      if(ii.eq.1.or.jj.eq.1) go to 1300
      if(jj.lt.iwini(ii) .or. jj.gt.iwlst(ii)) stop 630
      iw = jj-iwini(ii)+1
      ispc = iwstr(ii)
      if(ispc.gt.0) then
         k = instr(iw,ii)
      elseif(ispc.lt.0) then
         jv = -ispc
         k = ilstp(iw,jv)
      else
         stop 634
      endif
      if(k.lt.0) iflp = iflp+1
      i = ii
      j = jj
      go to 1200
 1300 continue
      if(ii.ne.1.or.jj.ne.1) stop 640
      g(1) = 0.0d0
      if(ifly.gt.0) then
         pfly = (dble(ifly)/dble(iflw))*(100.0d0)
      else
         pfly = 0.0d0
      endif
      if(iflp.gt.0) then
         pflp = (dble(iflp)/dble(iflw))*(100.0d0)
      else
         pflp = 0.0d0
      endif
c
      return
      end
c
*INTOPT
c
c     subroutine intopt
c
c     for approximating L2 norm of q1(t) - sqrt(d/dt g(t)) x q2(g(t))
c     from grid point (ii,jj) to grid point (i,j) in the unit squrare,
c     where q1 and q2 are in the L2 space of functions of norm 1
c     from [0,1] into the plane, and g is the straight line joining
c     the two grid points.
c
      subroutine intopt(q1, q2, qm1, qm2, qc, t1, t2, s, r, iar,
     *                  area, mgrid, ii, jj, i, j, iopt)
c
      double precision qm1(*), qm2(*)
      double precision q1(2,*), q2(2,*), qc(2,*)
      double precision t1(*), t2(*), s(*), r(*), area
      integer iar(*)
      integer ii, jj, i, j, j2, mgrid, iopt
c
      double precision slope, slops, y
      integer ix, k, ilen
c
      if(ii.ge.i .or. jj.ge.j) stop 650
c
c     compute slope of line joining grid points
c
      slope = (t2(j)-t2(jj))/(t1(i)-t1(ii))
      slops = dsqrt(slope)
c
c     compute q1(t)-sqrt(slope) x q2(g(t))
c
      if(iopt.eq.1) go to 1000
      ix = i
      ilen = 2
  200 continue
      ix = iar(ix)
      if(ix.eq.0) stop 660
      ilen = ilen+1
      if(ix.ne.ii) go to 200
      ilen = ilen-1
c
      qc(1, 1) = q1(1,ii)-slops*q2(1,jj)
      qc(2, 1) = q1(2,ii)-slops*q2(2,jj)
      qc(1,ilen) = q1(1, i)-slops*q2(1, j)
      qc(2,ilen) = q1(2, i)-slops*q2(2, j)
c
      s(1) = t1(ii)
      s(ilen) = t1(i) 
c
      if(ilen.eq.2) go to 600
      ix = i
      k = ilen
  500 continue
      ix = iar(ix)
      if(ix.eq.ii) go to 600
      k = k-1
      y = (t1(ix)-t1(ii))*slope + t2(jj)
      j2 = idnint(y*dble(mgrid-1)) + 1
      qc(1,k) = q1(1,ix)-slops*qm1(j2)
      qc(2,k) = q1(2,ix)-slops*qm2(j2)
      s(k) = t1(ix)
      go to 500
  600 continue
c
c     approximate L2 norm of q1(t)-sqrt(slope) x q2(g(t))
c
      call innpr2(qc, qc, s, r, ilen, area)
c
      return
c
 1000 continue
      ix = i
      ilen = 2
 1200 continue
      ix = iar(ix)
      if(ix.eq.0) stop 670
      ilen = ilen+1
      if(ix.ne.ii) go to 1200
      ilen = ilen-1
c
      qc(2, 1) = q1(2,ii)-slops*q2(2,jj)
      qc(2,ilen) = q1(2, i)-slops*q2(2, j)
c
      s(1) = t1(ii)
      s(ilen) = t1(i) 
c
      if(ilen.eq.2) go to 1600
      ix = i
      k = ilen
 1500 continue
      ix = iar(ix)
      if(ix.eq.ii) go to 1600
      k = k-1
      y = (t1(ix)-t1(ii))*slope + t2(jj)
      j2 = idnint(y*dble(mgrid-1)) + 1
      qc(2,k) = q1(2,ix)-slops*qm2(j2)
      s(k) = t1(ix)
      go to 1500
 1600 continue
c
c     approximate L2 norm of q1(t)-sqrt(slope) x q2(g(t))
c
      call innpr1(qc, qc, s, r, ilen, area)
c
      return
      end
c
*INNPR2
c
c     subroutine innpr2 to
c
c     approximate area under a curve defined by Tm1 points
c     whose y-coordinates are the result of computing the
c     inner product of Tm1 pairs of points; its x-coordinates
c     are given as input
c
c     2-d case
c
      subroutine innpr2(q1, q2, x, y, Tm1, area)
c
      double precision q1(2,*), q2(2,*), x(*), y(*), area
      integer Tm1, i
c
      do 100 i = 1, Tm1
         y(i) = q1(1,i)*q2(1,i) + q1(2,i)*q2(2,i)
  100 continue
c
      call trapz2(x, y, Tm1, area)
c
      return
      end
c
*INNPR1
c
c     subroutine innpr1 to
c
c     approximate area under a curve defined by Tm1 points
c     whose y-coordinates are the result of computing the
c     inner product of Tm1 pairs of points; its x-coordinates
c     are given as input
c
c     1-d case
c
      subroutine innpr1(q1, q2, x, y, Tm1, area)
c
      double precision q1(2,*), q2(2,*), x(*), y(*), area
      integer Tm1, i
c
      do 100 i = 1, Tm1
         y(i) = q1(2,i)*q2(2,i)
  100 continue
c
      call trapz2(x, y, Tm1, area)
c
      return
      end
c
*TRAPZ2
c
c     subroutine trapz2 to
c
c     approximate area under a curve using trapezoidal approximation;
c     arrays x and y contain x- and y-coordinates of points on curve
c
      subroutine trapz2(x, y, Tm1, area)
c
      double precision x(*), y(*), area
      integer Tm1, i
c
      area = 0.0d0
      do 100 i = 2, Tm1
         area = area + (y(i)+y(i-1))*(x(i)-x(i-1))
  100 continue
      area = area/2.0d0
c
      return
      end
c
*PCHSP
      SUBROUTINE PCHSP(IC,VC,N,X,F,D,INCFD,WK,NWK,IERR)
C***BEGIN PROLOGUE  PCHSP
C***DATE WRITTEN   820503   (YYMMDD)
C***REVISION DATE  870707   (YYMMDD)
C***CATEGORY NO.  E1B
C***KEYWORDS  LIBRARY=SLATEC(PCHIP),
C             TYPE=SINGLE PRECISION(PCHSP-S DPCHSP-D),
C             CUBIC HERMITE INTERPOLATION,PIECEWISE CUBIC INTERPOLATION,
C             SPLINE INTERPOLATION
C***AUTHOR  FRITSCH, F. N., (LLNL)
C             MATHEMATICS AND STATISTICS DIVISION
C             LAWRENCE LIVERMORE NATIONAL LABORATORY
C             P.O. BOX 808  (L-316)
C             LIVERMORE, CA  94550
C             FTS 532-4275, (415) 422-4275
C***PURPOSE  Set derivatives needed to determine the Hermite represen-
C            tation of the cubic spline interpolant to given data, with
C            specified boundary conditions.
C***DESCRIPTION
C
C          PCHSP:   Piecewise Cubic Hermite Spline
C
C     Computes the Hermite representation of the cubic spline inter-
C     polant to the data given in X and F satisfying the boundary
C     conditions specified by IC and VC.
C
C     To facilitate two-dimensional applications, includes an increment
C     between successive values of the F- and D-arrays.
C
C     The resulting piecewise cubic Hermite function may be evaluated
C     by PCHFE or PCHFD.
C
C     NOTE:  This is a modified version of C. de Boor'S cubic spline
C            routine CUBSPL.
C
C ----------------------------------------------------------------------
C
C  Calling sequence:
C
C        PARAMETER  (INCFD = ...)
C        INTEGER  IC(2), N, NWK, IERR
C        REAL  VC(2), X(N), F(INCFD,N), D(INCFD,N), WK(NWK)
C
C        CALL  PCHSP (IC, VC, N, X, F, D, INCFD, WK, NWK, IERR)
C
C   Parameters:
C
C     IC -- (input) integer array of length 2 specifying desired
C           boundary conditions:
C           IC(1) = IBEG, desired condition at beginning of data.
C           IC(2) = IEND, desired condition at end of data.
C
C           IBEG = 0  to set D(1) so that the third derivative is con-
C              tinuous at X(2).  This is the "not a knot" condition
C              provided by de Boor'S cubic spline routine CUBSPL.
C              < This is the default boundary condition. >
C           IBEG = 1  if first derivative at X(1) is given in VC(1).
C           IBEG = 2  if second derivative at X(1) is given in VC(1).
C           IBEG = 3  to use the 3-point difference formula for D(1).
C                     (Reverts to the default b.c. if N.LT.3 .)
C           IBEG = 4  to use the 4-point difference formula for D(1).
C                     (Reverts to the default b.c. if N.LT.4 .)
C          NOTES:
C           1. An error return is taken if IBEG is out of range.
C           2. For the "natural" boundary condition, use IBEG=2 and
C              VC(1)=0.
C
C           IEND may take on the same values as IBEG, but applied to
C           derivative at X(N).  In case IEND = 1 or 2, the value is
C           given in VC(2).
C
C          NOTES:
C           1. An error return is taken if IEND is out of range.
C           2. For the "natural" boundary condition, use IEND=2 and
C              VC(2)=0.
C
C     VC -- (input) real array of length 2 specifying desired boundary
C           values, as indicated above.
C           VC(1) need be set only if IC(1) = 1 or 2 .
C           VC(2) need be set only if IC(2) = 1 or 2 .
C
C     N -- (input) number of data points.  (Error return if N.LT.2 .)
C
C     X -- (input) real array of independent variable values.  The
C           elements of X must be strictly increasing:
C                X(I-1) .LT. X(I),  I = 2(1)N.
C           (Error return if not.)
C
C     F -- (input) real array of dependent variable values to be inter-
C           polated.  F(1+(I-1)*INCFD) is value corresponding to X(I).
C
C     D -- (output) real array of derivative values at the data points.
C           These values will determine the cubic spline interpolant
C           with the requested boundary conditions.
C           The value corresponding to X(I) is stored in
C                D(1+(I-1)*INCFD),  I=1(1)N.
C           No other entries in D are changed.
C
C     INCFD -- (input) increment between successive values in F and D.
C           This argument is provided primarily for 2-D applications.
C           (Error return if  INCFD.LT.1 .)
C
C     WK -- (scratch) real array of working storage.
C
C     NWK -- (input) length of work array.
C           (Error return if NWK.LT.2*N .)
C
C     IERR -- (output) error flag.
C           Normal return:
C              IERR = 0  (no errors).
C           "Recoverable" errors:
C              IERR = -1  if N.LT.2 .
C              IERR = -2  if INCFD.LT.1 .
C              IERR = -3  if the X-array is not strictly increasing.
C              IERR = -4  if IBEG.LT.0 or IBEG.GT.4 .
C              IERR = -5  if IEND.LT.0 of IEND.GT.4 .
C              IERR = -6  if both of the above are true.
C              IERR = -7  if NWK is too small.
C               NOTE:  The above errors are checked in the order listed,
C                   and following arguments have **NOT** been validated.
C             (The D-array has not been changed in any of these cases.)
C              IERR = -8  in case of trouble solving the linear system
C                         for the interior derivative values.
C             (The D-array may have been changed in this case.)
C             (             Do **NOT** use it!                )
C
C***REFERENCES  CARL DE BOOR, A PRACTICAL GUIDE TO SPLINES, SPRINGER-
C                 VERLAG (NEW YORK, 1978), PP. 53-59.
C***ROUTINES CALLED  PCHDF,XERROR
C***END PROLOGUE  PCHSP
C
C ----------------------------------------------------------------------
C
C  Change record:
C     82-08-04   Converted to SLATEC library version.
C     87-07-07   Minor cosmetic changes to prologue.
C
C ----------------------------------------------------------------------
C
C  Programming notes:
C
C     To produce a double precision version, simply:
C        a. Change PCHSP to DPCHSP wherever it occurs,
C        b. Change the real declarations to double precision, and
C        c. Change the constants ZERO, HALF, ... to double precision.
C
C  DECLARE ARGUMENTS.
C
      INTEGER  IC(2), N, INCFD, NWK, IERR
      double precision VC(2), X(N), F(INCFD,N), D(INCFD,N), WK(2,N)
C
C  DECLARE LOCAL VARIABLES.
C
      INTEGER  IBEG, IEND, INDEX, J, NM1
      double precision G,HALF,ONE, STEMP(3), THREE, TWO, XTEMP(4), ZERO
      double precision PCHDF
C
      DATA  ZERO /0./,  HALF /0.5/,  ONE /1./,  TWO /2./,  THREE /3./
C
C  VALIDITY-CHECK ARGUMENTS.
C
C***FIRST EXECUTABLE STATEMENT  PCHSP
      IF ( N.LT.2 )  GO TO 5001
      IF ( INCFD.LT.1 )  GO TO 5002
      DO 1  J = 2, N
         IF ( X(J).LE.X(J-1) )  GO TO 5003
    1 CONTINUE
C
      IBEG = IC(1)
      IEND = IC(2)
      IERR = 0
      IF ( (IBEG.LT.0).OR.(IBEG.GT.4) )  IERR = IERR - 1
      IF ( (IEND.LT.0).OR.(IEND.GT.4) )  IERR = IERR - 2
      IF ( IERR.LT.0 )  GO TO 5004
C
C  FUNCTION DEFINITION IS OK -- GO ON.
C
      IF ( NWK .LT. 2*N )  GO TO 5007
C
C  COMPUTE FIRST DIFFERENCES OF X SEQUENCE AND STORE IN WK(1,.). ALSO,
C  COMPUTE FIRST DIVIDED DIFFERENCE OF DATA AND STORE IN WK(2,.).
      DO 5  J=2,N
         WK(1,J) = X(J) - X(J-1)
         WK(2,J) = (F(1,J) - F(1,J-1))/WK(1,J)
    5 CONTINUE
C
C  SET TO DEFAULT BOUNDARY CONDITIONS IF N IS TOO SMALL.
C
      IF ( IBEG.GT.N )  IBEG = 0
      IF ( IEND.GT.N )  IEND = 0
C
C  SET UP FOR BOUNDARY CONDITIONS.
C
      IF ( (IBEG.EQ.1).OR.(IBEG.EQ.2) )  THEN
         D(1,1) = VC(1)
      ELSE IF (IBEG .GT. 2)  THEN
C        PICK UP FIRST IBEG POINTS, IN REVERSE ORDER.
         DO 10  J = 1, IBEG
            INDEX = IBEG-J+1
C           INDEX RUNS FROM IBEG DOWN TO 1.
            XTEMP(J) = X(INDEX)
            IF (J .LT. IBEG)  STEMP(J) = WK(2,INDEX)
   10    CONTINUE
C                 --------------------------------
         D(1,1) = PCHDF (IBEG, XTEMP, STEMP, IERR)
C                 --------------------------------
         IF (IERR .NE. 0)  GO TO 5009
         IBEG = 1
      ENDIF
C
      IF ( (IEND.EQ.1).OR.(IEND.EQ.2) )  THEN
         D(1,N) = VC(2)
      ELSE IF (IEND .GT. 2)  THEN
C        PICK UP LAST IEND POINTS.
         DO 15  J = 1, IEND
            INDEX = N-IEND+J
C           INDEX RUNS FROM N+1-IEND UP TO N.
            XTEMP(J) = X(INDEX)
            IF (J .LT. IEND)  STEMP(J) = WK(2,INDEX+1)
   15    CONTINUE
C                 --------------------------------
         D(1,N) = PCHDF (IEND, XTEMP, STEMP, IERR)
C                 --------------------------------
         IF (IERR .NE. 0)  GO TO 5009
         IEND = 1
      ENDIF
C
C --------------------( BEGIN CODING FROM CUBSPL )--------------------
C
C  **** A TRIDIAGONAL LINEAR SYSTEM FOR THE UNKNOWN SLOPES S(J) OF
C  F  AT X(J), J=1,...,N, IS GENERATED AND THEN SOLVED BY GAUSS ELIM-
C  INATION, WITH S(J) ENDING UP IN D(1,J), ALL J.
C     WK(1,.) AND WK(2,.) ARE USED FOR TEMPORARY STORAGE.
C
C  CONSTRUCT FIRST EQUATION FROM FIRST BOUNDARY CONDITION, OF THE FORM
C             WK(2,1)*S(1) + WK(1,1)*S(2) = D(1,1)
C
      IF (IBEG .EQ. 0)  THEN
         IF (N .EQ. 2)  THEN
C           NO CONDITION AT LEFT END AND N = 2.
            WK(2,1) = ONE
            WK(1,1) = ONE
            D(1,1) = TWO*WK(2,2)
         ELSE
C           NOT-A-KNOT CONDITION AT LEFT END AND N .GT. 2.
            WK(2,1) = WK(1,3)
            WK(1,1) = WK(1,2) + WK(1,3)
            D(1,1) =((WK(1,2) + TWO*WK(1,1))*WK(2,2)*WK(1,3)
     *                        + WK(1,2)**2*WK(2,3)) / WK(1,1)
         ENDIF
      ELSE IF (IBEG .EQ. 1)  THEN
C        SLOPE PRESCRIBED AT LEFT END.
         WK(2,1) = ONE
         WK(1,1) = ZERO
      ELSE
C        SECOND DERIVATIVE PRESCRIBED AT LEFT END.
         WK(2,1) = TWO
         WK(1,1) = ONE
         D(1,1) = THREE*WK(2,2) - HALF*WK(1,2)*D(1,1)
      ENDIF
C
C  IF THERE ARE INTERIOR KNOTS, GENERATE THE CORRESPONDING EQUATIONS AND
C  CARRY OUT THE FORWARD PASS OF GAUSS ELIMINATION, AFTER WHICH THE J-TH
C  EQUATION READS    WK(2,J)*S(J) + WK(1,J)*S(J+1) = D(1,J).
C
      NM1 = N-1
      IF (NM1 .GT. 1)  THEN
         DO 20 J=2,NM1
            IF (WK(2,J-1) .EQ. ZERO)  GO TO 5008
            G = -WK(1,J+1)/WK(2,J-1)
            D(1,J) = G*D(1,J-1)
     *                  + THREE*(WK(1,J)*WK(2,J+1) + WK(1,J+1)*WK(2,J))
            WK(2,J) = G*WK(1,J-1) + TWO*(WK(1,J) + WK(1,J+1))
   20    CONTINUE
      ENDIF
C
C  CONSTRUCT LAST EQUATION FROM SECOND BOUNDARY CONDITION, OF THE FORM
C           (-G*WK(2,N-1))*S(N-1) + WK(2,N)*S(N) = D(1,N)
C
C     IF SLOPE IS PRESCRIBED AT RIGHT END, ONE CAN GO DIRECTLY TO BACK-
C     SUBSTITUTION, SINCE ARRAYS HAPPEN TO BE SET UP JUST RIGHT FOR IT
C     AT THIS POINT.
      IF (IEND .EQ. 1)  GO TO 30
C
      IF (IEND .EQ. 0)  THEN
         IF (N.EQ.2 .AND. IBEG.EQ.0)  THEN
C           NOT-A-KNOT AT RIGHT ENDPOINT AND AT LEFT ENDPOINT AND N = 2.
            D(1,2) = WK(2,2)
            GO TO 30
         ELSE IF ((N.EQ.2) .OR. (N.EQ.3 .AND. IBEG.EQ.0))  THEN
C           EITHER (N=3 AND NOT-A-KNOT ALSO AT LEFT) OR (N=2 AND *NOT*
C           NOT-A-KNOT AT LEFT END POINT).
            D(1,N) = TWO*WK(2,N)
            WK(2,N) = ONE
            IF (WK(2,N-1) .EQ. ZERO)  GO TO 5008
            G = -ONE/WK(2,N-1)
         ELSE
C           NOT-A-KNOT AND N .GE. 3, AND EITHER N.GT.3 OR  ALSO NOT-A-
C           KNOT AT LEFT END POINT.
            G = WK(1,N-1) + WK(1,N)
C           DO NOT NEED TO CHECK FOLLOWING DENOMINATORS (X-DIFFERENCES).
            D(1,N) = ((WK(1,N)+TWO*G)*WK(2,N)*WK(1,N-1)
     *                  + WK(1,N)**2*(F(1,N-1)-F(1,N-2))/WK(1,N-1))/G
            IF (WK(2,N-1) .EQ. ZERO)  GO TO 5008
            G = -G/WK(2,N-1)
            WK(2,N) = WK(1,N-1)
         ENDIF
      ELSE
C        SECOND DERIVATIVE PRESCRIBED AT RIGHT ENDPOINT.
         D(1,N) = THREE*WK(2,N) + HALF*WK(1,N)*D(1,N)
         WK(2,N) = TWO
         IF (WK(2,N-1) .EQ. ZERO)  GO TO 5008
         G = -ONE/WK(2,N-1)
      ENDIF
C
C  COMPLETE FORWARD PASS OF GAUSS ELIMINATION.
C
      WK(2,N) = G*WK(1,N-1) + WK(2,N)
      IF (WK(2,N) .EQ. ZERO)   GO TO 5008
      D(1,N) = (G*D(1,N-1) + D(1,N))/WK(2,N)
C
C  CARRY OUT BACK SUBSTITUTION
C
   30 CONTINUE
      DO 40 J=NM1,1,-1
         IF (WK(2,J) .EQ. ZERO)  GO TO 5008
         D(1,J) = (D(1,J) - WK(1,J)*D(1,J+1))/WK(2,J)
   40 CONTINUE
C --------------------(  END  CODING FROM CUBSPL )--------------------
C
C  NORMAL RETURN.
C
      RETURN
C
C  ERROR RETURNS.
C
 5001 CONTINUE
C     N.LT.2 RETURN.
      IERR = -1
c     CALL XERROR ('PCHSP -- NUMBER OF DATA POINTS LESS THAN TWO'
c    *           , 44, IERR, 1)
      write(*,*)'pchsp error: number of data points less than two'
      RETURN
C
 5002 CONTINUE
C     INCFD.LT.1 RETURN.
      IERR = -2
c     CALL XERROR ('PCHSP -- INCREMENT LESS THAN ONE'
c    *           , 32, IERR, 1)
      write(*,*)'pchsp error: increment less than one'
      RETURN
C
 5003 CONTINUE
C     X-ARRAY NOT STRICTLY INCREASING.
      IERR = -3
c     CALL XERROR ('PCHSP -- X-ARRAY NOT STRICTLY INCREASING'
c    *           , 40, IERR, 1)
      write(*,*)'pchsp error: x-array not strictly increasing'
      RETURN
C
 5004 CONTINUE
C     IC OUT OF RANGE RETURN.
      IERR = IERR - 3
c     CALL XERROR ('PCHSP -- IC OUT OF RANGE'
c    *           , 24, IERR, 1)
      write(*,*)'pchsp error: ic out of range'
      RETURN
C
 5007 CONTINUE
C     NWK TOO SMALL RETURN.
      IERR = -7
c     CALL XERROR ('PCHSP -- WORK ARRAY TOO SMALL'
c    *           , 29, IERR, 1)
      write(*,*)'pchsp error: work array too small'
      RETURN
C
 5008 CONTINUE
C     SINGULAR SYSTEM.
C   *** THEORETICALLY, THIS CAN ONLY OCCUR IF SUCCESSIVE X-VALUES   ***
C   *** ARE EQUAL, WHICH SHOULD ALREADY HAVE BEEN CAUGHT (IERR=-3). ***
      IERR = -8
c     CALL XERROR ('PCHSP -- SINGULAR LINEAR SYSTEM'
c    *           , 31, IERR, 1)
      write(*,*)'pchsp error: singular linear system'
      RETURN
C
 5009 CONTINUE
C     ERROR RETURN FROM PCHDF.
C   *** THIS CASE SHOULD NEVER OCCUR ***
      IERR = -9
c     CALL XERROR ('PCHSP -- ERROR RETURN FROM PCHDF'
c    *           , 32, IERR, 1)
      write(*,*)'pchsp error: error return from pchdf'
      RETURN
C------------- LAST LINE OF PCHSP FOLLOWS ------------------------------
      END
*DECK PCHDF
      double precision FUNCTION PCHDF (K, X, S, IERR)
C***BEGIN PROLOGUE  PCHDF
C***SUBSIDIARY
C***PURPOSE  Computes divided differences for PCHCE and PCHSP
C***LIBRARY   SLATEC (PCHIP)
C***TYPE      SINGLE PRECISION (PCHDF-S, DPCHDF-D)
C***AUTHOR  Fritsch, F. N., (LLNL)
C***DESCRIPTION
C
C          PCHDF:   PCHIP Finite Difference Formula
C
C     Uses a divided difference formulation to compute a K-point approx-
C     imation to the derivative at X(K) based on the data in X and S.
C
C     Called by  PCHCE  and  PCHSP  to compute 3- and 4-point boundary
C     derivative approximations.
C
C ----------------------------------------------------------------------
C
C     On input:
C        K      is the order of the desired derivative approximation.
C               K must be at least 3 (error return if not).
C        X      contains the K values of the independent variable.
C               X need not be ordered, but the values **MUST** be
C               distinct.  (Not checked here.)
C        S      contains the associated slope values:
C                  S(I) = (F(I+1)-F(I))/(X(I+1)-X(I)), I=1(1)K-1.
C               (Note that S need only be of length K-1.)
C
C     On return:
C        S      will be destroyed.
C        IERR   will be set to -1 if K.LT.2 .
C        PCHDF  will be set to the desired derivative approximation if
C               IERR=0 or to zero if IERR=-1.
C
C ----------------------------------------------------------------------
C
C***SEE ALSO  PCHCE, PCHSP
C***REFERENCES  Carl de Boor, A Practical Guide to Splines, Springer-
C                 Verlag, New York, 1978, pp. 10-16.
C***ROUTINES CALLED  XERMSG
C***REVISION HISTORY  (YYMMDD)
C   820503  DATE WRITTEN
C   820805  Converted to SLATEC library version.
C   870813  Minor cosmetic changes.
C   890411  Added SAVE statements (Vers. 3.2).
C   890411  REVISION DATE from Version 3.2
C   891214  Prologue converted to Version 4.0 format.  (BAB)
C   900315  CALLs to XERROR changed to CALLs to XERMSG.  (THJ)
C   900328  Added TYPE section.  (WRB)
C   910408  Updated AUTHOR and DATE WRITTEN sections in prologue.  (WRB)
C   920429  Revised format and order of references.  (WRB,FNF)
C   930503  Improved purpose.  (FNF)
C***END PROLOGUE  PCHDF
C
C**End
C
C  DECLARE ARGUMENTS.
C
      INTEGER  K, IERR
      double precision X(K), S(K)
C
C  DECLARE LOCAL VARIABLES.
C
      INTEGER  I, J
      double precision VALUE, ZERO
      SAVE ZERO
      DATA  ZERO /0./
C
C  CHECK FOR LEGAL VALUE OF K.
C
C***FIRST EXECUTABLE STATEMENT  PCHDF
      IF (K .LT. 3)  GO TO 5001
C
C  COMPUTE COEFFICIENTS OF INTERPOLATING POLYNOMIAL.
C
      DO 10  J = 2, K-1
         DO 9  I = 1, K-J
            S(I) = (S(I+1)-S(I))/(X(I+J)-X(I))
    9    CONTINUE
   10 CONTINUE
C
C  EVALUATE DERIVATIVE AT X(K).
C
      VALUE = S(1)
      DO 20  I = 2, K-1
         VALUE = S(I) + VALUE*(X(K)-X(I))
   20 CONTINUE
C
C  NORMAL RETURN.
C
      IERR = 0
      PCHDF = VALUE
      RETURN
C
C  ERROR RETURN.
C
 5001 CONTINUE
C     K.LT.3 RETURN.
      IERR = -1
c     CALL XERMSG ('SLATEC', 'PCHDF', 'K LESS THAN THREE', IERR, 1)
      write(*,*)'slatec pchdf error: k less than three'
      PCHDF = ZERO
      RETURN
C------------- LAST LINE OF PCHDF FOLLOWS ------------------------------
      END
      SUBROUTINE PCHFE(N,X,F,D,INCFD,SKIP,NE,XE,FE,IERR)
C***BEGIN PROLOGUE  PCHFE
C***DATE WRITTEN   811020   (YYMMDD)
C***REVISION DATE  870707   (YYMMDD)
C***CATEGORY NO.  E3
C***KEYWORDS  LIBRARY=SLATEC(PCHIP),
C             TYPE=SINGLE PRECISION(PCHFE-S DPCHFE-D),
C             CUBIC HERMITE EVALUATION,HERMITE INTERPOLATION,
C             PIECEWISE CUBIC EVALUATION
C***AUTHOR  FRITSCH, F. N., (LLNL)
C             MATHEMATICS AND STATISTICS DIVISION
C             LAWRENCE LIVERMORE NATIONAL LABORATORY
C             P.O. BOX 808  (L-316)
C             LIVERMORE, CA  94550
C             FTS 532-4275, (415) 422-4275
C***PURPOSE  Evaluate a piecewise cubic Hermite function at an array of
C            points.  May be used by itself for Hermite interpolation,
C            or as an evaluator for PCHIM or PCHIC.
C***DESCRIPTION
C
C          PCHFE:  Piecewise Cubic Hermite Function Evaluator
C
C     Evaluates the cubic Hermite function defined by  N, X, F, D  at
C     the points  XE(J), J=1(1)NE.
C
C     To provide compatibility with PCHIM and PCHIC, includes an
C     increment between successive values of the F- and D-arrays.
C
C ----------------------------------------------------------------------
C
C  Calling sequence:
C
C        PARAMETER  (INCFD = ...)
C        INTEGER  N, NE, IERR
C        REAL  X(N), F(INCFD,N), D(INCFD,N), XE(NE), FE(NE)
C        LOGICAL  SKIP
C
C        CALL  PCHFE (N, X, F, D, INCFD, SKIP, NE, XE, FE, IERR)
C
C   Parameters:
C
C     N -- (input) number of data points.  (Error return if N.LT.2 .)
C
C     X -- (input) real array of independent variable values.  The
C           elements of X must be strictly increasing:
C                X(I-1) .LT. X(I),  I = 2(1)N.
C           (Error return if not.)
C
C     F -- (input) real array of function values.  F(1+(I-1)*INCFD) is
C           the value corresponding to X(I).
C
C     D -- (input) real array of derivative values.  D(1+(I-1)*INCFD) is
C           the value corresponding to X(I).
C
C     INCFD -- (input) increment between successive values in F and D.
C           (Error return if  INCFD.LT.1 .)
C
C     SKIP -- (input/output) logical variable which should be set to
C           .TRUE. if the user wishes to skip checks for validity of
C           preceding parameters, or to .FALSE. otherwise.
C           This will save time in case these checks have already
C           been performed (say, in PCHIM or PCHIC).
C           SKIP will be set to .TRUE. on normal return.
C
C     NE -- (input) number of evaluation points.  (Error return if
C           NE.LT.1 .)
C
C     XE -- (input) real array of points at which the function is to be
C           evaluated.
C
C          NOTES:
C           1. The evaluation will be most efficient if the elements
C              of XE are increasing relative to X;
C              that is,   XE(J) .GE. X(I)
C              implies    XE(K) .GE. X(I),  all K.GE.J .
C           2. If any of the XE are outside the interval [X(1),X(N)],
C              values are extrapolated from the nearest extreme cubic,
C              and a warning error is returned.
C
C     FE -- (output) real array of values of the cubic Hermite function
C           defined by  N, X, F, D  at the points  XE.
C
C     IERR -- (output) error flag.
C           Normal return:
C              IERR = 0  (no errors).
C           Warning error:
C              IERR.GT.0  means that extrapolation was performed at
C                 IERR points.
C           "Recoverable" errors:
C              IERR = -1  if N.LT.2 .
C              IERR = -2  if INCFD.LT.1 .
C              IERR = -3  if the X-array is not strictly increasing.
C              IERR = -4  if NE.LT.1 .
C             (The FE-array has not been changed in any of these cases.)
C               NOTE:  The above errors are checked in the order listed,
C                   and following arguments have **NOT** been validated.
C
C***REFERENCES  (NONE)
C***ROUTINES CALLED  CHFEV,XERROR
C***END PROLOGUE  PCHFE
C
C ----------------------------------------------------------------------
C
C  Change record:
C     82-08-03   Minor cosmetic changes for release 1.
C     87-07-07   Minor cosmetic changes to prologue.
C
C ----------------------------------------------------------------------
C
C  Programming notes:
C
C     1. To produce a double precision version, simply:
C        a. Change PCHFE to DPCHFE, and CHFEV to DCHFEV, wherever they
C           occur,
C        b. Change the real declaration to double precision,
C
C     2. Most of the coding between the call to CHFEV and the end of
C        the IR-loop could be eliminated if it were permissible to
C        assume that XE is ordered relative to X.
C
C     3. CHFEV does not assume that X1 is less than X2.  thus, it would
C        be possible to write a version of PCHFE that assumes a strict-
C        ly decreasing X-array by simply running the IR-loop backwards
C        (and reversing the order of appropriate tests).
C
C     4. The present code has a minor bug, which I have decided is not
C        worth the effort that would be required to fix it.
C        If XE contains points in [X(N-1),X(N)], followed by points .LT.
C        X(N-1), followed by points .GT.X(N), the extrapolation points
C        will be counted (at least) twice in the total returned in IERR.
C
C  DECLARE ARGUMENTS.
C
      INTEGER  N, INCFD, NE, IERR
      double precision X(N), F(INCFD,N), D(INCFD,N), XE(NE), FE(NE)
      LOGICAL  SKIP
C
C  DECLARE LOCAL VARIABLES.
C
      INTEGER  I, IERC, IR, J, JFIRST, NEXT(2), NJ
C
C  VALIDITY-CHECK ARGUMENTS.
C
C***FIRST EXECUTABLE STATEMENT  PCHFE
      IF (SKIP)  GO TO 5
C
      IF ( N.LT.2 )  GO TO 5001
      IF ( INCFD.LT.1 )  GO TO 5002
      DO 1  I = 2, N
         IF ( X(I).LE.X(I-1) )  GO TO 5003
    1 CONTINUE
C
C  FUNCTION DEFINITION IS OK, GO ON.
C
    5 CONTINUE
      IF ( NE.LT.1 )  GO TO 5004
      IERR = 0
      SKIP = .TRUE.
C
C  LOOP OVER INTERVALS.        (   INTERVAL INDEX IS  IL = IR-1  . )
C                              ( INTERVAL IS X(IL).LE.X.LT.X(IR) . )
      JFIRST = 1
      IR = 2
   10 CONTINUE
C
C     SKIP OUT OF LOOP IF HAVE PROCESSED ALL EVALUATION POINTS.
C
         IF (JFIRST .GT. NE)  GO TO 5000
C
C     LOCATE ALL POINTS IN INTERVAL.
C
         DO 20  J = JFIRST, NE
            IF (XE(J) .GE. X(IR))  GO TO 30
   20    CONTINUE
         J = NE + 1
         GO TO 40
C
C     HAVE LOCATED FIRST POINT BEYOND INTERVAL.
C
   30    CONTINUE
         IF (IR .EQ. N)  J = NE + 1
C
   40    CONTINUE
         NJ = J - JFIRST
C
C     SKIP EVALUATION IF NO POINTS IN INTERVAL.
C
         IF (NJ .EQ. 0)  GO TO 50
C
C     EVALUATE CUBIC AT XE(I),  I = JFIRST (1) J-1 .
C
C       ----------------------------------------------------------------
        CALL CHFEV (X(IR-1),X(IR), F(1,IR-1),F(1,IR), D(1,IR-1),D(1,IR),
     *              NJ, XE(JFIRST), FE(JFIRST), NEXT, IERC)
C       ----------------------------------------------------------------
         IF (IERC .LT. 0)  GO TO 5005
C
         IF (NEXT(2) .EQ. 0)  GO TO 42
C        IF (NEXT(2) .GT. 0)  THEN
C           IN THE CURRENT SET OF XE-POINTS, THERE ARE NEXT(2) TO THE
C           RIGHT OF X(IR).
C
            IF (IR .LT. N)  GO TO 41
C           IF (IR .EQ. N)  THEN
C              THESE ARE ACTUALLY EXTRAPOLATION POINTS.
               IERR = IERR + NEXT(2)
               GO TO 42
   41       CONTINUE
C           ELSE
C              WE SHOULD NEVER HAVE GOTTEN HERE.
               GO TO 5005
C           ENDIF
C        ENDIF
   42    CONTINUE
C
         IF (NEXT(1) .EQ. 0)  GO TO 49
C        IF (NEXT(1) .GT. 0)  THEN
C           IN THE CURRENT SET OF XE-POINTS, THERE ARE NEXT(1) TO THE
C           LEFT OF X(IR-1).
C
            IF (IR .GT. 2)  GO TO 43
C           IF (IR .EQ. 2)  THEN
C              THESE ARE ACTUALLY EXTRAPOLATION POINTS.
               IERR = IERR + NEXT(1)
               GO TO 49
   43       CONTINUE
C           ELSE
C              XE IS NOT ORDERED RELATIVE TO X, SO MUST ADJUST
C              EVALUATION INTERVAL.
C
C              FIRST, LOCATE FIRST POINT TO LEFT OF X(IR-1).
               DO 44  I = JFIRST, J-1
                  IF (XE(I) .LT. X(IR-1))  GO TO 45
   44          CONTINUE
C              NOTE-- CANNOT DROP THROUGH HERE UNLESS THERE IS AN ERROR
C                     IN CHFEV.
               GO TO 5005
C
   45          CONTINUE
C              RESET J.  (THIS WILL BE THE NEW JFIRST.)
               J = I
C
C              NOW FIND OUT HOW FAR TO BACK UP IN THE X-ARRAY.
               DO 46  I = 1, IR-1
                  IF (XE(J) .LT. X(I)) GO TO 47
   46          CONTINUE
C              NB-- CAN NEVER DROP THROUGH HERE, SINCE XE(J).LT.X(IR-1).
C
   47          CONTINUE
C              AT THIS POINT, EITHER  XE(J) .LT. X(1)
C                 OR      X(I-1) .LE. XE(J) .LT. X(I) .
C              RESET IR, RECOGNIZING THAT IT WILL BE INCREMENTED BEFORE
C              CYCLING.
               IR = MAX0(1, I-1)
C           ENDIF
C        ENDIF
   49    CONTINUE
C
         JFIRST = J
C
C     END OF IR-LOOP.
C
   50 CONTINUE
      IR = IR + 1
      IF (IR .LE. N)  GO TO 10
C
C  NORMAL RETURN.
C
 5000 CONTINUE
      RETURN
C
C  ERROR RETURNS.
C
 5001 CONTINUE
C     N.LT.2 RETURN.
      IERR = -1
c     CALL XERROR ('PCHFE -- NUMBER OF DATA POINTS LESS THAN TWO'
c    *           , 44, IERR, 1)
      write(*,*)'pchfe error: number of data points less than two'
      RETURN
C
 5002 CONTINUE
C     INCFD.LT.1 RETURN.
      IERR = -2
c     CALL XERROR ('PCHFE -- INCREMENT LESS THAN ONE'
c    *           , 32, IERR, 1)
      write(*,*)'pchfe error: increment less than one'
      RETURN
C
 5003 CONTINUE
C     X-ARRAY NOT STRICTLY INCREASING.
      IERR = -3
c     CALL XERROR ('PCHFE -- X-ARRAY NOT STRICTLY INCREASING'
c    *           , 40, IERR, 1)
      write(*,*)'pchfe error: x-array not strictly increasing'
      RETURN
C
 5004 CONTINUE
C     NE.LT.1 RETURN.
      IERR = -4
c     CALL XERROR ('PCHFE -- NUMBER OF EVALUATION POINTS LESS THAN ONE'
c    *           , 50, IERR, 1)
      write(*,*)'pchfe error: number of evaluation points less than one'
      RETURN
C
 5005 CONTINUE
C     ERROR RETURN FROM CHFEV.
C   *** THIS CASE SHOULD NEVER OCCUR ***
      IERR = -5
c     CALL XERROR ('PCHFE -- ERROR RETURN FROM CHFEV -- FATAL'
c    *           , 41, IERR, 2)
      write(*,*)'pchfe error: error return form chfev -- fatal'
      RETURN
C------------- LAST LINE OF PCHFE FOLLOWS ------------------------------
      END
      SUBROUTINE CHFEV(X1,X2,F1,F2,D1,D2,NE,XE,FE,NEXT,IERR)
C***BEGIN PROLOGUE  CHFEV
C***DATE WRITTEN   811019   (YYMMDD)
C***REVISION DATE  870707   (YYMMDD)
C***CATEGORY NO.  E3,H1
C***KEYWORDS  LIBRARY=SLATEC(PCHIP),
C             TYPE=SINGLE PRECISION(CHFEV-S DCHFEV-D),
C             CUBIC HERMITE EVALUATION,CUBIC POLYNOMIAL EVALUATION
C***AUTHOR  FRITSCH, F. N., (LLNL)
C             MATHEMATICS AND STATISTICS DIVISION
C             LAWRENCE LIVERMORE NATIONAL LABORATORY
C             P.O. BOX 808  (L-316)
C             LIVERMORE, CA  94550
C             FTS 532-4275, (415) 422-4275
C***PURPOSE  Evaluate a cubic polynomial given in Hermite form at an
C            array of points.  While designed for use by PCHFE, it may
C            be useful directly as an evaluator for a piecewise cubic
C            Hermite function in applications, such as graphing, where
C            the interval is known in advance.
C***DESCRIPTION
C
C          CHFEV:  Cubic Hermite Function EValuator
C
C     Evaluates the cubic polynomial determined by function values
C     F1,F2 and derivatives D1,D2 on interval (X1,X2) at the points
C     XE(J), J=1(1)NE.
C
C ----------------------------------------------------------------------
C
C  Calling sequence:
C
C        INTEGER  NE, NEXT(2), IERR
C        REAL  X1, X2, F1, F2, D1, D2, XE(NE), FE(NE)
C
C        CALL  CHFEV (X1,X2, F1,F2, D1,D2, NE, XE, FE, NEXT, IERR)
C
C   Parameters:
C
C     X1,X2 -- (input) endpoints of interval of definition of cubic.
C           (Error return if  X1.EQ.X2 .)
C
C     F1,F2 -- (input) values of function at X1 and X2, respectively.
C
C     D1,D2 -- (input) values of derivative at X1 and X2, respectively.
C
C     NE -- (input) number of evaluation points.  (Error return if
C           NE.LT.1 .)
C
C     XE -- (input) real array of points at which the function is to be
C           evaluated.  If any of the XE are outside the interval
C           [X1,X2], a warning error is returned in NEXT.
C
C     FE -- (output) real array of values of the cubic function defined
C           by  X1,X2, F1,F2, D1,D2  at the points  XE.
C
C     NEXT -- (output) integer array indicating number of extrapolation
C           points:
C            NEXT(1) = number of evaluation points to left of interval.
C            NEXT(2) = number of evaluation points to right of interval.
C
C     IERR -- (output) error flag.
C           Normal return:
C              IERR = 0  (no errors).
C           "Recoverable" errors:
C              IERR = -1  if NE.LT.1 .
C              IERR = -2  if X1.EQ.X2 .
C                (The FE-array has not been changed in either case.)
C
C***REFERENCES  (NONE)
C***ROUTINES CALLED  XERROR
C***END PROLOGUE  CHFEV
C
C ----------------------------------------------------------------------
C
C  Change record:
C     82-08-03   Minor cosmetic changes for release 1.
C
C ----------------------------------------------------------------------
C
C  Programming notes:
C
C     To produce a double precision version, simply:
C        a. Change CHFEV to DCHFEV wherever it occurs,
C        b. Change the real declaration to double precision,
C        c. Change the constant ZERO to double precision, and
C        d. Change the names of the Fortran functions:  AMAX1, AMIN1.
C
C  DECLARE ARGUMENTS.
C
      INTEGER  NE, NEXT(2), IERR
      double precision X1, X2, F1, F2, D1, D2, XE(NE), FE(NE)
C
C  DECLARE LOCAL VARIABLES.
C
      INTEGER  I
      double precision C2, C3, DEL1, DEL2, DELTA, H, X, XMI, XMA, ZERO
      DATA  ZERO /0./
C
C  VALIDITY-CHECK ARGUMENTS.
C
C***FIRST EXECUTABLE STATEMENT  CHFEV
      IF (NE .LT. 1)  GO TO 5001
      H = X2 - X1
      IF (H .EQ. ZERO)  GO TO 5002
C
C  INITIALIZE.
C
      IERR = 0
      NEXT(1) = 0
      NEXT(2) = 0
      XMI = DMIN1(ZERO, H)
      XMA = DMAX1(ZERO, H)
C
C  COMPUTE CUBIC COEFFICIENTS (EXPANDED ABOUT X1).
C
      DELTA = (F2 - F1)/H
      DEL1 = (D1 - DELTA)/H
      DEL2 = (D2 - DELTA)/H
C                                           (DELTA IS NO LONGER NEEDED.)
      C2 = -(DEL1+DEL1 + DEL2)
      C3 = (DEL1 + DEL2)/H
C                               (H, DEL1 AND DEL2 ARE NO LONGER NEEDED.)
C
C  EVALUATION LOOP.
C
      DO 500  I = 1, NE
         X = XE(I) - X1
         FE(I) = F1 + X*(D1 + X*(C2 + X*C3))
C          COUNT EXTRAPOLATION POINTS.
         IF ( X.LT.XMI )  NEXT(1) = NEXT(1) + 1
         IF ( X.GT.XMA )  NEXT(2) = NEXT(2) + 1
C        (NOTE REDUNDANCY--IF EITHER CONDITION IS TRUE, OTHER IS FALSE.)
  500 CONTINUE
C
C  NORMAL RETURN.
C
      RETURN
C
C  ERROR RETURNS.
C
 5001 CONTINUE
C     NE.LT.1 RETURN.
      IERR = -1
c     CALL XERROR ('CHFEV -- NUMBER OF EVALUATION POINTS LESS THAN ONE'
c    *           , 50, IERR, 1)
      write(*,*)'chfev error: number of evaluation points less than one'
      RETURN
C
 5002 CONTINUE
C     X1.EQ.X2 RETURN.
      IERR = -2
c     CALL XERROR ('CHFEV -- INTERVAL ENDPOINTS EQUAL'
c    *           , 33, IERR, 1)
      write(*,*)'chfev error: interval endpoints equal'
      RETURN
C------------- LAST LINE OF CHFEV FOLLOWS ------------------------------
      END
